@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,42 +15,27 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """The ListView to display downloads in."""
 
 import functools
+from typing import Callable, MutableSequence, Tuple, Union
 
-from PyQt5.QtCore import pyqtSlot, QSize, Qt, QTimer
+from PyQt5.QtCore import pyqtSlot, QSize, Qt
 from PyQt5.QtWidgets import QListView, QSizePolicy, QMenu, QStyleFactory
 
 from qutebrowser.browser import downloads
-from qutebrowser.config import config
-from qutebrowser.utils import qtutils, utils, objreg
-from qutebrowser.qt import sip
+from qutebrowser.config import stylesheet
+from qutebrowser.utils import qtutils, utils
 
 
-def update_geometry(obj):
-    """Weird WORKAROUND for some weird PyQt bug (probably).
-
-    This actually should be a method of DownloadView, but for some reason the
-    rowsInserted/rowsRemoved signals don't get disconnected from this method
-    when the DownloadView is deleted from Qt (e.g. by closing a window).
-
-    Here we check if obj ("self") was deleted and just ignore the event if so.
-
-    Original bug:   https://github.com/qutebrowser/qutebrowser/issues/167
-    Workaround bug: https://github.com/qutebrowser/qutebrowser/issues/171
-    """
-    def _update_geometry():
-        """Actually update the geometry if the object still exists."""
-        if sip.isdeleted(obj):
-            return
-        obj.updateGeometry()
-
-    # If we don't use a singleShot QTimer, the geometry isn't updated correctly
-    # and won't include the new item.
-    QTimer.singleShot(0, _update_geometry)
+_ActionListType = MutableSequence[
+    Union[
+        Tuple[None, None],  # separator
+        Tuple[str, Callable[[], None]],
+    ]
+]
 
 
 class DownloadView(QListView):
@@ -59,7 +44,6 @@ class DownloadView(QListView):
 
     Attributes:
         _menu: The QMenu which is currently displayed.
-        _model: The currently set model.
     """
 
     STYLESHEET = """
@@ -73,11 +57,11 @@ class DownloadView(QListView):
         }
     """
 
-    def __init__(self, win_id, parent=None):
+    def __init__(self, model, parent=None):
         super().__init__(parent)
         if not utils.is_mac:
             self.setStyle(QStyleFactory.create('Fusion'))
-        config.set_register_stylesheet(self)
+        stylesheet.set_register(self)
         self.setResizeMode(QListView.Adjust)
         self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
         self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
@@ -85,10 +69,9 @@ class DownloadView(QListView):
         self.setFlow(QListView.LeftToRight)
         self.setSpacing(1)
         self._menu = None
-        model = objreg.get('download-model', scope='window', window=win_id)
-        model.rowsInserted.connect(functools.partial(update_geometry, self))
-        model.rowsRemoved.connect(functools.partial(update_geometry, self))
-        model.dataChanged.connect(functools.partial(update_geometry, self))
+        model.rowsInserted.connect(self._update_geometry)
+        model.rowsRemoved.connect(self._update_geometry)
+        model.dataChanged.connect(self._update_geometry)
         self.setModel(model)
         self.setWrapping(True)
         self.setContextMenuPolicy(Qt.CustomContextMenu)
@@ -98,11 +81,20 @@ class DownloadView(QListView):
     def __repr__(self):
         model = self.model()
         if model is None:
-            count = 'None'
+            count = 'None'  # type: ignore[unreachable]
         else:
             count = model.rowCount()
         return utils.get_repr(self, count=count)
 
+    @pyqtSlot()
+    def _update_geometry(self):
+        """Wrapper to call updateGeometry.
+
+        For some reason, this is needed so that PyQt disconnects the signals and handles
+        arguments correctly. Probably a WORKAROUND for an unknown PyQt bug.
+        """
+        self.updateGeometry()
+
     @pyqtSlot(bool)
     def on_fullscreen_requested(self, on):
         """Hide/show the downloadview when entering/leaving fullscreen."""
@@ -125,29 +117,31 @@ class DownloadView(QListView):
             item.open_file()
             item.remove()
 
-    def _get_menu_actions(self, item):
+    def _get_menu_actions(
+            self,
+            item: downloads.AbstractDownloadItem
+    ) -> _ActionListType:
         """Get the available context menu actions for a given DownloadItem.
 
         Args:
             item: The DownloadItem to get the actions for, or None.
-
-        Return:
-            A list of either:
-                - (QAction, callable) tuples.
-                - (None, None) for a separator
         """
         model = self.model()
-        actions = []
+        actions: _ActionListType = []
         if item is None:
             pass
         elif item.done:
             if item.successful:
                 actions.append(("Open", item.open_file))
+                actions.append(("Open directory", functools.partial(
+                    item.open_file, open_dir=True, cmdline=None)))
             else:
                 actions.append(("Retry", item.try_retry))
             actions.append(("Remove", item.remove))
         else:
             actions.append(("Cancel", item.cancel))
+        actions.append(("Copy URL", functools.partial(
+            utils.set_clipboard, item.url().toDisplayString())))
         if model.can_clear():
             actions.append((None, None))
             actions.append(("Remove all finished", model.download_clear))
@@ -167,6 +161,8 @@ class DownloadView(QListView):
             if name is None and handler is None:
                 self._menu.addSeparator()
             else:
+                assert name is not None
+                assert handler is not None
                 action = self._menu.addAction(name)
                 action.triggered.connect(handler)
         if actions:
