@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,21 +15,22 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """A HintManager to draw hints over links."""
 
 import collections
-import typing
 import functools
 import os
 import re
 import html
 import enum
+import dataclasses
 from string import ascii_lowercase
+from typing import (TYPE_CHECKING, Callable, Dict, Iterable, Iterator, List, Mapping,
+                    MutableSequence, Optional, Sequence, Set)
 
-import attr
-from PyQt5.QtCore import pyqtSlot, QObject, Qt, QUrl
+from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject, Qt, QUrl
 from PyQt5.QtWidgets import QLabel
 
 from qutebrowser.config import config, configexc
@@ -38,14 +39,30 @@ from qutebrowser.browser import webelem, history
 from qutebrowser.commands import userscripts, runners
 from qutebrowser.api import cmdutils
 from qutebrowser.utils import usertypes, log, qtutils, message, objreg, utils
-if typing.TYPE_CHECKING:
+if TYPE_CHECKING:
     from qutebrowser.browser import browsertab
 
 
-Target = enum.Enum('Target', ['normal', 'current', 'tab', 'tab_fg', 'tab_bg',
-                              'window', 'yank', 'yank_primary', 'run', 'fill',
-                              'hover', 'download', 'userscript', 'spawn',
-                              'delete', 'right_click'])
+class Target(enum.Enum):
+
+    """What action to take on a hint."""
+
+    normal = enum.auto()
+    current = enum.auto()
+    tab = enum.auto()
+    tab_fg = enum.auto()
+    tab_bg = enum.auto()
+    window = enum.auto()
+    yank = enum.auto()
+    yank_primary = enum.auto()
+    run = enum.auto()
+    fill = enum.auto()
+    hover = enum.auto()
+    download = enum.auto()
+    userscript = enum.auto()
+    spawn = enum.auto()
+    delete = enum.auto()
+    right_click = enum.auto()
 
 
 class HintingError(Exception):
@@ -133,7 +150,7 @@ class HintLabel(QLabel):
         self.deleteLater()
 
 
-@attr.s
+@dataclasses.dataclass
 class HintContext:
 
     """Context namespace used for hinting.
@@ -164,22 +181,23 @@ class HintContext:
         group: The group of web elements to hint.
     """
 
-    all_labels = attr.ib(attr.Factory(list))  # type: typing.List[HintLabel]
-    labels = attr.ib(attr.Factory(dict))  # type: typing.Dict[str, HintLabel]
-    target = attr.ib(None)  # type: Target
-    baseurl = attr.ib(None)  # type: QUrl
-    to_follow = attr.ib(None)  # type: str
-    rapid = attr.ib(False)  # type: bool
-    first_run = attr.ib(True)  # type: bool
-    add_history = attr.ib(False)  # type: bool
-    filterstr = attr.ib(None)  # type: str
-    args = attr.ib(attr.Factory(list))  # type: typing.List[str]
-    tab = attr.ib(None)  # type: browsertab.AbstractTab
-    group = attr.ib(None)  # type: str
-    hint_mode = attr.ib(None)  # type: str
-    first = attr.ib(False)  # type: bool
+    tab: 'browsertab.AbstractTab'
+    target: Target
+    rapid: bool
+    hint_mode: str
+    add_history: bool
+    first: bool
+    baseurl: QUrl
+    args: List[str]
+    group: str
 
-    def get_args(self, urlstr: str) -> typing.Sequence[str]:
+    all_labels: List[HintLabel] = dataclasses.field(default_factory=list)
+    labels: Dict[str, HintLabel] = dataclasses.field(default_factory=dict)
+    to_follow: Optional[str] = None
+    first_run: bool = True
+    filterstr: Optional[str] = None
+
+    def get_args(self, urlstr: str) -> Sequence[str]:
         """Get the arguments, with {hint-url} replaced by the given URL."""
         args = []
         for arg in self.args:
@@ -336,8 +354,8 @@ class HintActions:
         commandrunner.run_safely('spawn ' + ' '.join(args))
 
 
-_ElemsType = typing.Sequence[webelem.AbstractWebElement]
-_HintStringsType = typing.MutableSequence[str]
+_ElemsType = Sequence[webelem.AbstractWebElement]
+_HintStringsType = MutableSequence[str]
 
 
 class HintManager(QObject):
@@ -353,7 +371,7 @@ class HintManager(QObject):
         _tab_id: The tab ID this HintManager is associated with.
 
     Signals:
-        See HintActions
+        set_text: Request for the statusbar to change its text.
     """
 
     HINT_TEXTS = {
@@ -375,11 +393,13 @@ class HintManager(QObject):
         Target.delete: "Delete an element",
     }
 
+    set_text = pyqtSignal(str)
+
     def __init__(self, win_id: int, parent: QObject = None) -> None:
         """Constructor."""
         super().__init__(parent)
         self._win_id = win_id
-        self._context = None  # type: typing.Optional[HintContext]
+        self._context: Optional[HintContext] = None
         self._word_hinter = WordHinter()
 
         self._actions = HintActions(win_id)
@@ -402,10 +422,8 @@ class HintManager(QObject):
         for label in self._context.all_labels:
             label.cleanup()
 
-        text = self._get_text()
-        message_bridge = objreg.get('message-bridge', scope='window',
-                                    window=self._win_id)
-        message_bridge.maybe_reset_text(text)
+        self.set_text.emit('')
+
         self._context = None
 
     def _hint_strings(self, elems: _ElemsType) -> _HintStringsType:
@@ -511,12 +529,10 @@ class HintManager(QObject):
         Return:
             A list of shuffled hint strings.
         """
-        buckets = [
-            [] for i in range(length)
-        ]  # type: typing.Sequence[_HintStringsType]
+        buckets: Sequence[_HintStringsType] = [[] for i in range(length)]
         for i, hint in enumerate(hints):
             buckets[i % len(buckets)].append(hint)
-        result = []  # type: _HintStringsType
+        result: _HintStringsType = []
         for bucket in buckets:
             result += bucket
         return result
@@ -541,7 +557,7 @@ class HintManager(QObject):
             A hint string.
         """
         base = len(chars)
-        hintstr = []  # type: typing.MutableSequence[str]
+        hintstr: MutableSequence[str] = []
         remainder = 0
         while True:
             remainder = number % base
@@ -575,7 +591,7 @@ class HintManager(QObject):
                 raise cmdutils.CommandError(
                     "'args' is only allowed with target userscript/spawn.")
 
-    def _filter_matches(self, filterstr: str, elemstr: str) -> bool:
+    def _filter_matches(self, filterstr: Optional[str], elemstr: str) -> bool:
         """Return True if `filterstr` matches `elemstr`."""
         # Empty string and None always match
         if not filterstr:
@@ -633,12 +649,11 @@ class HintManager(QObject):
         keyparser = self._get_keyparser(usertypes.KeyMode.hint)
         keyparser.update_bindings(strings)
 
-        message_bridge = objreg.get('message-bridge', scope='window',
-                                    window=self._win_id)
-        message_bridge.set_text(self._get_text())
         modeman.enter(self._win_id, usertypes.KeyMode.hint,
                       'HintManager.start')
 
+        self.set_text.emit(self._get_text())
+
         if self._context.first:
             self._fire(strings[0])
             return
@@ -744,19 +759,23 @@ class HintManager(QObject):
                                             "with target {}!".format(name))
 
         self._check_args(target, *args)
-        self._context = HintContext()
-        self._context.tab = tab
-        self._context.target = target
-        self._context.rapid = rapid
-        self._context.hint_mode = self._get_hint_mode(mode)
-        self._context.add_history = add_history
-        self._context.first = first
+
         try:
-            self._context.baseurl = tabbed_browser.current_url()
+            baseurl = tabbed_browser.current_url()
         except qtutils.QtValueError:
             raise cmdutils.CommandError("No URL set for this page yet!")
-        self._context.args = list(args)
-        self._context.group = group
+
+        self._context = HintContext(
+            tab=tab,
+            target=target,
+            rapid=rapid,
+            hint_mode=self._get_hint_mode(mode),
+            add_history=add_history,
+            first=first,
+            baseurl=baseurl,
+            args=list(args),
+            group=group,
+        )
 
         try:
             selector = webelem.css_selector(self._context.group,
@@ -770,7 +789,7 @@ class HintManager(QObject):
             error_cb=lambda err: message.error(str(err)),
             only_visible=True)
 
-    def _get_hint_mode(self, mode: typing.Optional[str]) -> str:
+    def _get_hint_mode(self, mode: Optional[str]) -> str:
         """Get the hinting mode to use based on a mode argument."""
         if mode is None:
             return config.val.hints.mode
@@ -782,7 +801,7 @@ class HintManager(QObject):
             raise cmdutils.CommandError("Invalid mode: {}".format(e))
         return mode
 
-    def current_mode(self) -> typing.Optional[str]:
+    def current_mode(self) -> Optional[str]:
         """Return the currently active hinting mode (or None otherwise)."""
         if self._context is None:
             return None
@@ -793,7 +812,7 @@ class HintManager(QObject):
             self,
             keystr: str = "",
             filterstr: str = "",
-            visible: typing.Mapping[str, HintLabel] = None
+            visible: Mapping[str, HintLabel] = None
     ) -> None:
         """Handle the auto_follow option."""
         assert self._context is not None
@@ -819,7 +838,7 @@ class HintManager(QObject):
         else:
             follow = False
             # save the keystr of the only one visible hint to be picked up
-            # later by self.follow_hint
+            # later by self.hint_follow
             self._context.to_follow = list(visible.keys())[0]
 
         if follow:
@@ -855,7 +874,7 @@ class HintManager(QObject):
                 pass
         self._handle_auto_follow(keystr=keystr)
 
-    def filter_hints(self, filterstr: typing.Optional[str]) -> None:
+    def filter_hints(self, filterstr: Optional[str]) -> None:
         """Filter displayed hints according to a text.
 
         Args:
@@ -981,8 +1000,8 @@ class HintManager(QObject):
             self._context.first_run = False
 
     @cmdutils.register(instance='hintmanager', scope='window',
-                       modes=[usertypes.KeyMode.hint])
-    def follow_hint(self, select: bool = False, keystring: str = None) -> None:
+                       modes=[usertypes.KeyMode.hint], deprecated_name='follow-hint')
+    def hint_follow(self, select: bool = False, keystring: str = None) -> None:
         """Follow a hint.
 
         Args:
@@ -1026,7 +1045,7 @@ class WordHinter:
 
     def __init__(self) -> None:
         # will be initialized on first use.
-        self.words = set()  # type: typing.Set[str]
+        self.words: Set[str] = set()
         self.dictionary = None
 
     def ensure_initialized(self) -> None:
@@ -1052,16 +1071,16 @@ class WordHinter:
                             hints.discard(word[:i + 1])
                         hints.add(word)
                     self.words.update(hints)
-            except IOError as e:
+            except OSError as e:
                 error = "Word hints requires reading the file at {}: {}"
                 raise HintingError(error.format(dictionary, str(e)))
 
     def extract_tag_words(
             self, elem: webelem.AbstractWebElement
-    ) -> typing.Iterator[str]:
+    ) -> Iterator[str]:
         """Extract tag words form the given element."""
-        _extractor_type = typing.Callable[[webelem.AbstractWebElement], str]
-        attr_extractors = {
+        _extractor_type = Callable[[webelem.AbstractWebElement], str]
+        attr_extractors: Mapping[str, _extractor_type] = {
             "alt": lambda elem: elem["alt"],
             "name": lambda elem: elem["name"],
             "title": lambda elem: elem["title"],
@@ -1069,7 +1088,7 @@ class WordHinter:
             "src": lambda elem: elem["src"].split('/')[-1],
             "href": lambda elem: elem["href"].split('/')[-1],
             "text": str,
-        }  # type: typing.Mapping[str, _extractor_type]
+        }
 
         extractable_attrs = collections.defaultdict(list, {
             "img": ["alt", "title", "src"],
@@ -1085,8 +1104,8 @@ class WordHinter:
 
     def tag_words_to_hints(
             self,
-            words: typing.Iterable[str]
-    ) -> typing.Iterator[str]:
+            words: Iterable[str]
+    ) -> Iterator[str]:
         """Take words and transform them to proper hints if possible."""
         for candidate in words:
             if not candidate:
@@ -1097,20 +1116,20 @@ class WordHinter:
             if 4 < match.end() - match.start() < 8:
                 yield candidate[match.start():match.end()].lower()
 
-    def any_prefix(self, hint: str, existing: typing.Iterable[str]) -> bool:
+    def any_prefix(self, hint: str, existing: Iterable[str]) -> bool:
         return any(hint.startswith(e) or e.startswith(hint) for e in existing)
 
     def filter_prefixes(
             self,
-            hints: typing.Iterable[str],
-            existing: typing.Iterable[str]
-    ) -> typing.Iterator[str]:
+            hints: Iterable[str],
+            existing: Iterable[str]
+    ) -> Iterator[str]:
         """Filter hints which don't start with the given prefix."""
         return (h for h in hints if not self.any_prefix(h, existing))
 
     def new_hint_for(self, elem: webelem.AbstractWebElement,
-                     existing: typing.Iterable[str],
-                     fallback: typing.Iterable[str]) -> typing.Optional[str]:
+                     existing: Iterable[str],
+                     fallback: Iterable[str]) -> Optional[str]:
         """Return a hint for elem, not conflicting with the existing."""
         new = self.tag_words_to_hints(self.extract_tag_words(elem))
         new_no_prefixes = self.filter_prefixes(new, existing)
@@ -1134,7 +1153,7 @@ class WordHinter:
         """
         self.ensure_initialized()
         hints = []
-        used_hints = set()  # type: typing.Set[str]
+        used_hints: Set[str] = set()
         words = iter(self.words)
         for elem in elems:
             hint = self.new_hint_for(elem, used_hints, words)
