@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 # Copyright 2015 Daniel Schadt
 #
 # This file is part of qutebrowser.
@@ -16,7 +16,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """pdf.js integration for qutebrowser."""
 
@@ -24,12 +24,23 @@ import os
 
 from PyQt5.QtCore import QUrl, QUrlQuery
 
-from qutebrowser.utils import (utils, javascript, jinja, qtutils, usertypes,
-                               standarddir, log)
-from qutebrowser.misc import objects
+from qutebrowser.utils import utils, javascript, jinja, standarddir, log
 from qutebrowser.config import config
 
 
+_SYSTEM_PATHS = [
+    # Debian pdf.js-common
+    # Arch Linux pdfjs (AUR)
+    '/usr/share/pdf.js/',
+    # Flatpak (Flathub)
+    '/app/share/pdf.js/',
+    # Arch Linux pdf.js (AUR)
+    '/usr/share/javascript/pdf.js/',
+    # Debian libjs-pdf
+    '/usr/share/javascript/pdf/',
+]
+
+
 class PDFJSNotFound(Exception):
 
     """Raised when no pdf.js installation is found.
@@ -84,33 +95,24 @@ def _generate_pdfjs_script(filename):
     url_query.addQueryItem('filename', filename)
     url.setQuery(url_query)
 
+    js_url = javascript.to_js(
+        url.toString(QUrl.FullyEncoded))  # type: ignore[arg-type]
+
     return jinja.js_environment.from_string("""
         document.addEventListener("DOMContentLoaded", function() {
           if (typeof window.PDFJS !== 'undefined') {
               // v1.x
-              {% if disable_create_object_url %}
-              window.PDFJS.disableCreateObjectURL = true;
-              {% endif %}
               window.PDFJS.verbosity = window.PDFJS.VERBOSITY_LEVELS.info;
           } else {
               // v2.x
               const options = window.PDFViewerApplicationOptions;
-              {% if disable_create_object_url %}
-              options.set('disableCreateObjectURL', true);
-              {% endif %}
               options.set('verbosity', pdfjsLib.VerbosityLevel.INFOS);
           }
 
           const viewer = window.PDFView || window.PDFViewerApplication;
           viewer.open({{ url }});
         });
-    """).render(
-        url=javascript.to_js(url.toString(QUrl.FullyEncoded)),
-        # WORKAROUND for https://bugreports.qt.io/browse/QTBUG-70420
-        disable_create_object_url=(
-            not qtutils.version_check('5.12') and
-            not qtutils.version_check('5.7.1', exact=True, compiled=False) and
-            objects.backend == usertypes.Backend.QtWebEngine))
+    """).render(url=js_url)
 
 
 def get_pdfjs_res_and_path(path):
@@ -127,16 +129,7 @@ def get_pdfjs_res_and_path(path):
     content = None
     file_path = None
 
-    system_paths = [
-        # Debian pdf.js-common
-        # Arch Linux pdfjs (AUR)
-        '/usr/share/pdf.js/',
-        # Flatpak (Flathub)
-        '/app/share/pdf.js/',
-        # Arch Linux pdf.js (AUR)
-        '/usr/share/javascript/pdf.js/',
-        # Debian libjs-pdf
-        '/usr/share/javascript/pdf/',
+    system_paths = _SYSTEM_PATHS + [
         # fallback
         os.path.join(standarddir.data(), 'pdfjs'),
         # hardcoded fallback for --temp-basedir
@@ -156,7 +149,7 @@ def get_pdfjs_res_and_path(path):
     if content is None:
         res_path = '3rdparty/pdfjs/{}'.format(path)
         try:
-            content = utils.read_file(res_path, binary=True)
+            content = utils.read_file_binary(res_path)
         except FileNotFoundError:
             raise PDFJSNotFound(path) from None
         except OSError as e:
@@ -221,6 +214,7 @@ def is_available():
     """Return true if a pdfjs installation is available."""
     try:
         get_pdfjs_res('build/pdf.js')
+        get_pdfjs_res('web/viewer.html')
     except PDFJSNotFound:
         return False
     else:
@@ -237,11 +231,13 @@ def should_use_pdfjs(mimetype, url):
     return is_pdf and not is_download_url and config_enabled
 
 
-def get_main_url(filename):
+def get_main_url(filename: str, original_url: QUrl) -> QUrl:
     """Get the URL to be opened to view a local PDF."""
     url = QUrl('qute://pdfjs/web/viewer.html')
     query = QUrlQuery()
     query.addQueryItem('filename', filename)  # read from our JS
     query.addQueryItem('file', '')  # to avoid pdfjs opening the default PDF
+    urlstr = original_url.toString(QUrl.FullyEncoded)  # type: ignore[arg-type]
+    query.addQueryItem('source', urlstr)
     url.setQuery(query)
     return url
