@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2015-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2015-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,21 +15,20 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Customized QWebInspector for QtWebEngine."""
 
-import os
 import pathlib
 
-from PyQt5.QtCore import QUrl, QLibraryInfo
+from PyQt5.QtCore import QLibraryInfo
 from PyQt5.QtWebEngineWidgets import QWebEngineView, QWebEnginePage
 from PyQt5.QtWidgets import QWidget
 
 from qutebrowser.browser import inspector
 from qutebrowser.browser.webengine import webenginesettings
 from qutebrowser.misc import miscwidgets
-from qutebrowser.utils import version, qtutils
+from qutebrowser.utils import version
 
 
 class WebEngineInspectorView(QWebEngineView):
@@ -52,47 +51,9 @@ class WebEngineInspectorView(QWebEngineView):
         return self.page().inspectedPage().view().createWindow(wintype)
 
 
-def supports_new() -> bool:
-    """Check whether a new-style inspector is supported."""
-    return hasattr(QWebEnginePage, 'setInspectedPage')
-
-
-class LegacyWebEngineInspector(inspector.AbstractWebInspector):
-
-    """A web inspector for QtWebEngine without Qt API support.
-
-    Only needed with Qt <= 5.10.
-    """
-
-    def __init__(self, splitter: miscwidgets.InspectorSplitter,
-                 win_id: int,
-                 parent: QWidget = None) -> None:
-        super().__init__(splitter, win_id, parent)
-        self._ensure_enabled()
-        view = WebEngineInspectorView()
-        self._settings = webenginesettings.WebEngineSettings(view.settings())
-        self._set_widget(view)
-
-    def _ensure_enabled(self) -> None:
-        if 'QTWEBENGINE_REMOTE_DEBUGGING' not in os.environ:
-            raise inspector.Error(
-                "QtWebEngine inspector is not enabled. See "
-                "'qutebrowser --help' for details.")
-
-    def inspect(self, page: QWebEnginePage) -> None:  # type: ignore[override]
-        # We're lying about the URL here a bit, but this way, URL patterns for
-        # Qt 5.11/5.12/5.13 also work in this case.
-        self._settings.update_for_url(QUrl('chrome-devtools://devtools'))
-        port = int(os.environ['QTWEBENGINE_REMOTE_DEBUGGING'])
-        self._widget.load(QUrl('http://localhost:{}/'.format(port)))
-
-
 class WebEngineInspector(inspector.AbstractWebInspector):
 
-    """A web inspector for QtWebEngine with Qt API support.
-
-    Available since Qt 5.11.
-    """
+    """A web inspector for QtWebEngine with Qt API support."""
 
     def __init__(self, splitter: miscwidgets.InspectorSplitter,
                  win_id: int,
@@ -130,4 +91,4 @@ class WebEngineInspector(inspector.AbstractWebInspector):
 
         WORKAROUND for what's likely an unknown Qt bug.
         """
-        return qtutils.version_check('5.12')
+        return True
