@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,10 +15,11 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Configuration files residing on disk."""
 
+import enum
 import pathlib
 import types
 import os.path
@@ -27,8 +28,9 @@ import textwrap
 import traceback
 import configparser
 import contextlib
-import typing
 import re
+from typing import (TYPE_CHECKING, Any, Dict, Iterable, Iterator, List, Mapping,
+                    MutableMapping, Optional, cast)
 
 import yaml
 from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject, QSettings, qVersion
@@ -39,15 +41,42 @@ from qutebrowser.config import (configexc, config, configdata, configutils,
 from qutebrowser.keyinput import keyutils
 from qutebrowser.utils import standarddir, utils, qtutils, log, urlmatch
 
-if typing.TYPE_CHECKING:
+if TYPE_CHECKING:
     from qutebrowser.misc import savemanager
 
 
 # The StateConfig instance
-state = typing.cast('StateConfig', None)
+state = cast('StateConfig', None)
 
 
-_SettingsType = typing.Dict[str, typing.Dict[str, typing.Any]]
+_SettingsType = Dict[str, Dict[str, Any]]
+
+
+class VersionChange(enum.Enum):
+
+    """The type of version change when comparing two versions."""
+
+    unknown = enum.auto()
+    equal = enum.auto()
+    downgrade = enum.auto()
+
+    patch = enum.auto()
+    minor = enum.auto()
+    major = enum.auto()
+
+    def matches_filter(self, filterstr: str) -> bool:
+        """Whether the change matches a given filter.
+
+        This is intended to use filters like "major" (show major only), "minor" (show
+        major/minor) or "patch" (show all changes).
+        """
+        allowed_values: Dict[str, List[VersionChange]] = {
+            'major': [VersionChange.major],
+            'minor': [VersionChange.major, VersionChange.minor],
+            'patch': [VersionChange.major, VersionChange.minor, VersionChange.patch],
+            'never': [],
+        }
+        return self in allowed_values[filterstr]
 
 
 class StateConfig(configparser.ConfigParser):
@@ -58,15 +87,10 @@ class StateConfig(configparser.ConfigParser):
         super().__init__()
         self._filename = os.path.join(standarddir.data(), 'state')
         self.read(self._filename, encoding='utf-8')
-        qt_version = qVersion()
-        # We handle this here, so we can avoid setting qt_version_changed if
-        # the config is brand new, but can still set it when qt_version wasn't
-        # there before...
-        if 'general' in self:
-            old_qt_version = self['general'].get('qt_version', None)
-            self.qt_version_changed = old_qt_version != qt_version
-        else:
-            self.qt_version_changed = False
+
+        self.qt_version_changed = False
+        self.qutebrowser_version_changed = VersionChange.unknown
+        self._set_changed_attributes()
 
         for sect in ['general', 'geometry', 'inspector']:
             try:
@@ -77,14 +101,53 @@ class StateConfig(configparser.ConfigParser):
         deleted_keys = [
             ('general', 'fooled'),
             ('general', 'backend-warning-shown'),
+            ('general', 'old-qt-warning-shown'),
             ('geometry', 'inspector'),
         ]
         for sect, key in deleted_keys:
             self[sect].pop(key, None)
 
-        self['general']['qt_version'] = qt_version
+        self['general']['qt_version'] = qVersion()
         self['general']['version'] = qutebrowser.__version__
 
+    def _set_changed_attributes(self) -> None:
+        """Set qt_version_changed/qutebrowser_version_changed attributes.
+
+        We handle this here, so we can avoid setting qt_version_changed if
+        the config is brand new, but can still set it when qt_version wasn't
+        there before...
+        """
+        if 'general' not in self:
+            return
+
+        old_qt_version = self['general'].get('qt_version', None)
+        self.qt_version_changed = old_qt_version != qVersion()
+
+        old_qutebrowser_version = self['general'].get('version', None)
+        if old_qutebrowser_version is None:
+            # https://github.com/python/typeshed/issues/2093
+            return  # type: ignore[unreachable]
+
+        old_version = utils.parse_version(old_qutebrowser_version)
+        new_version = utils.parse_version(qutebrowser.__version__)
+
+        if old_version.isNull():
+            log.init.warning(f"Unable to parse old version {old_qutebrowser_version}")
+            return
+
+        assert not new_version.isNull(), qutebrowser.__version__
+
+        if old_version == new_version:
+            self.qutebrowser_version_changed = VersionChange.equal
+        elif new_version < old_version:
+            self.qutebrowser_version_changed = VersionChange.downgrade
+        elif old_version.segments()[:2] == new_version.segments()[:2]:
+            self.qutebrowser_version_changed = VersionChange.patch
+        elif old_version.majorVersion() == new_version.majorVersion():
+            self.qutebrowser_version_changed = VersionChange.minor
+        else:
+            self.qutebrowser_version_changed = VersionChange.major
+
     def init_save_manager(self,
                           save_manager: 'savemanager.SaveManager') -> None:
         """Make sure the config gets saved properly.
@@ -117,7 +180,7 @@ class YamlConfig(QObject):
                                       'autoconfig.yml')
         self._dirty = False
 
-        self._values = {}  # type: typing.Dict[str, configutils.Values]
+        self._values: Dict[str, configutils.Values] = {}
         for name, opt in configdata.DATA.items():
             self._values[name] = configutils.Values(opt)
 
@@ -130,7 +193,7 @@ class YamlConfig(QObject):
         """
         save_manager.add_saveable('yaml-config', self._save, self.changed)
 
-    def __iter__(self) -> typing.Iterator[configutils.Values]:
+    def __iter__(self) -> Iterator[configutils.Values]:
         """Iterate over configutils.Values items."""
         yield from self._values.values()
 
@@ -145,7 +208,7 @@ class YamlConfig(QObject):
         if not self._dirty:
             return
 
-        settings = {}  # type: _SettingsType
+        settings: _SettingsType = {}
         for name, values in sorted(self._values.items()):
             if not values:
                 continue
@@ -167,10 +230,7 @@ class YamlConfig(QObject):
             """.lstrip('\n')))
             utils.yaml_dump(data, f)
 
-    def _pop_object(self,
-                    yaml_data: typing.Any,
-                    key: str,
-                    typ: type) -> typing.Any:
+    def _pop_object(self, yaml_data: Any, key: str, typ: type) -> Any:
         """Get a global object from the given data."""
         if not isinstance(yaml_data, dict):
             desc = configexc.ConfigErrorDesc("While loading data",
@@ -227,19 +287,18 @@ class YamlConfig(QObject):
         self._validate_names(settings)
         self._build_values(settings)
 
-    def _load_settings_object(self, yaml_data: typing.Any) -> '_SettingsType':
+    def _load_settings_object(self, yaml_data: Any) -> '_SettingsType':
         """Load the settings from the settings: key."""
         return self._pop_object(yaml_data, 'settings', dict)
 
-    def _load_legacy_settings_object(self,
-                                     yaml_data: typing.Any) -> '_SettingsType':
+    def _load_legacy_settings_object(self, yaml_data: Any) -> '_SettingsType':
         data = self._pop_object(yaml_data, 'global', dict)
         settings = {}
         for name, value in data.items():
             settings[name] = {'global': value}
         return settings
 
-    def _build_values(self, settings: typing.Mapping) -> None:
+    def _build_values(self, settings: Mapping) -> None:
         """Build up self._values from the values in the given dict."""
         errors = []
         for name, yaml_values in settings.items():
@@ -285,7 +344,7 @@ class YamlConfig(QObject):
                       for e in sorted(unknown)]
             raise configexc.ConfigFileErrors('autoconfig.yml', errors)
 
-    def set_obj(self, name: str, value: typing.Any, *,
+    def set_obj(self, name: str, value: Any, *,
                 pattern: urlmatch.UrlPattern = None) -> None:
         """Set the given setting to the given value."""
         self._values[name].add(value, pattern)
@@ -477,8 +536,7 @@ class YamlMigrations(QObject):
                 self._settings[name][scope] = value
                 self.changed.emit()
 
-    def _migrate_to_multiple(self, old_name: str,
-                             new_names: typing.Iterable[str]) -> None:
+    def _migrate_to_multiple(self, old_name: str, new_names: Iterable[str]) -> None:
         if old_name not in self._settings:
             return
 
@@ -541,12 +599,12 @@ class ConfigAPI:
     def __init__(self, conf: config.Config, keyconfig: config.KeyConfig):
         self._config = conf
         self._keyconfig = keyconfig
-        self.errors = []  # type: typing.List[configexc.ConfigErrorDesc]
+        self.errors: List[configexc.ConfigErrorDesc] = []
         self.configdir = pathlib.Path(standarddir.config())
         self.datadir = pathlib.Path(standarddir.data())
 
     @contextlib.contextmanager
-    def _handle_error(self, action: str, name: str) -> typing.Iterator[None]:
+    def _handle_error(self, action: str, name: str) -> Iterator[None]:
         """Catch config-related exceptions and save them in self.errors."""
         try:
             yield
@@ -566,28 +624,35 @@ class ConfigAPI:
 
     def finalize(self) -> None:
         """Do work which needs to be done after reading config.py."""
+        if self._config.warn_autoconfig:
+            desc = configexc.ConfigErrorDesc(
+                "autoconfig loading not specified",
+                ("Your config.py should call either `config.load_autoconfig()`"
+                 " (to load settings configured via the GUI) or "
+                 "`config.load_autoconfig(False)` (to not do so)"))
+            self.errors.append(desc)
         self._config.update_mutables()
 
-    def load_autoconfig(self) -> None:
+    def load_autoconfig(self, load_config: bool = True) -> None:
         """Load the autoconfig.yml file which is used for :set/:bind/etc."""
-        with self._handle_error('reading', 'autoconfig.yml'):
-            read_autoconfig()
+        self._config.warn_autoconfig = False
+        if load_config:
+            with self._handle_error('reading', 'autoconfig.yml'):
+                read_autoconfig()
 
-    def get(self, name: str, pattern: str = None) -> typing.Any:
+    def get(self, name: str, pattern: str = None) -> Any:
         """Get a setting value from the config, optionally with a pattern."""
         with self._handle_error('getting', name):
             urlpattern = urlmatch.UrlPattern(pattern) if pattern else None
             return self._config.get_mutable_obj(name, pattern=urlpattern)
 
-    def set(self, name: str, value: typing.Any, pattern: str = None) -> None:
+    def set(self, name: str, value: Any, pattern: str = None) -> None:
         """Set a setting value in the config, optionally with a pattern."""
         with self._handle_error('setting', name):
             urlpattern = urlmatch.UrlPattern(pattern) if pattern else None
             self._config.set_obj(name, value, pattern=urlpattern)
 
-    def bind(self, key: str,
-             command: typing.Optional[str],
-             mode: str = 'normal') -> None:
+    def bind(self, key: str, command: Optional[str], mode: str = 'normal') -> None:
         """Bind a key to a command, with an optional key mode."""
         with self._handle_error('binding', key):
             seq = keyutils.KeySequence.parse(key)
@@ -616,7 +681,7 @@ class ConfigAPI:
             self.errors += e.errors
 
     @contextlib.contextmanager
-    def pattern(self, pattern: str) -> typing.Iterator[config.ConfigContainer]:
+    def pattern(self, pattern: str) -> Iterator[config.ConfigContainer]:
         """Get a ConfigContainer for the given pattern."""
         # We need to propagate the exception so we don't need to return
         # something.
@@ -632,9 +697,8 @@ class ConfigPyWriter:
 
     def __init__(
             self,
-            options: typing.List,
-            bindings: typing.MutableMapping[
-                str, typing.Mapping[str, typing.Optional[str]]],
+            options: List,
+            bindings: MutableMapping[str, Mapping[str, Optional[str]]],
             *, commented: bool) -> None:
         self._options = options
         self._bindings = bindings
@@ -655,7 +719,7 @@ class ConfigPyWriter:
         else:
             return line
 
-    def _gen_lines(self) -> typing.Iterator[str]:
+    def _gen_lines(self) -> Iterator[str]:
         """Generate a config.py with the given settings/bindings.
 
         Yields individual lines.
@@ -664,7 +728,7 @@ class ConfigPyWriter:
         yield from self._gen_options()
         yield from self._gen_bindings()
 
-    def _gen_header(self) -> typing.Iterator[str]:
+    def _gen_header(self) -> Iterator[str]:
         """Generate the initial header of the config."""
         yield self._line("# Autogenerated config.py")
         yield self._line("#")
@@ -689,15 +753,15 @@ class ConfigPyWriter:
                              "still loaded.")
             yield self._line("# Remove it to not load settings done via the "
                              "GUI.")
-            yield self._line("config.load_autoconfig()")
+            yield self._line("config.load_autoconfig(True)")
             yield ''
         else:
-            yield self._line("# Uncomment this to still load settings "
+            yield self._line("# Change the argument to True to still load settings "
                              "configured via autoconfig.yml")
-            yield self._line("# config.load_autoconfig()")
+            yield self._line("config.load_autoconfig(False)")
             yield ''
 
-    def _gen_options(self) -> typing.Iterator[str]:
+    def _gen_options(self) -> Iterator[str]:
         """Generate the options part of the config."""
         for pattern, opt, value in self._options:
             if opt.name in ['bindings.commands', 'bindings.default']:
@@ -725,7 +789,7 @@ class ConfigPyWriter:
                     opt.name, value, str(pattern)))
             yield ''
 
-    def _gen_bindings(self) -> typing.Iterator[str]:
+    def _gen_bindings(self) -> Iterator[str]:
         """Generate the bindings part of the config."""
         normal_bindings = self._bindings.pop('normal', {})
         if normal_bindings:
@@ -826,7 +890,7 @@ def read_autoconfig() -> None:
 
 
 @contextlib.contextmanager
-def saved_sys_properties() -> typing.Iterator[None]:
+def saved_sys_properties() -> Iterator[None]:
     """Save various sys properties such as sys.path and sys.modules."""
     old_path = sys.path.copy()
     old_modules = sys.modules.copy()
