@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,7 +15,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Other utilities which don't fit anywhere else."""
 
@@ -31,18 +31,30 @@ import traceback
 import functools
 import contextlib
 import posixpath
-import socket
 import shlex
-import glob
 import mimetypes
-import typing
+import pathlib
 import ctypes
 import ctypes.util
+from typing import (Any, Callable, IO, Iterator, Optional, Sequence, Tuple, Type, Union,
+                    Iterable, TYPE_CHECKING, cast)
+try:
+    # Protocol was added in Python 3.8
+    from typing import Protocol
+except ImportError:  # pragma: no cover
+    if not TYPE_CHECKING:
+        class Protocol:
 
-from PyQt5.QtCore import QUrl
-from PyQt5.QtGui import QColor, QClipboard, QDesktopServices
+            """Empty stub at runtime."""
+
+from PyQt5.QtCore import QUrl, QVersionNumber
+from PyQt5.QtGui import QClipboard, QDesktopServices
 from PyQt5.QtWidgets import QApplication
-import pkg_resources
+# We cannot use the stdlib version on 3.7-3.8 because we need the files() API.
+if sys.version_info >= (3, 9):
+    import importlib.resources as importlib_resources
+else:  # pragma: no cover
+    import importlib_resources
 import yaml
 try:
     from yaml import (CSafeLoader as YamlLoader,
@@ -54,7 +66,7 @@ except ImportError:  # pragma: no cover
     YAML_C_EXT = False
 
 import qutebrowser
-from qutebrowser.utils import qtutils, log
+from qutebrowser.utils import log
 
 
 fake_clipboard = None
@@ -67,6 +79,24 @@ is_windows = sys.platform.startswith('win')
 is_posix = os.name == 'posix'
 
 
+class SupportsLessThan(Protocol):
+
+    """Protocol for a "comparable" object."""
+
+    def __lt__(self, other: Any) -> bool:
+        ...
+
+
+if TYPE_CHECKING:
+    class VersionNumber(SupportsLessThan, QVersionNumber):
+
+        """WORKAROUND for incorrect PyQt stubs."""
+else:
+    class VersionNumber:
+
+        """We can't inherit from Protocol and QVersionNumber at runtime."""
+
+
 class Unreachable(Exception):
 
     """Raised when there was unreachable code."""
@@ -149,142 +179,87 @@ def compact_text(text: str, elidelength: int = None) -> str:
     return out
 
 
+def _resource_path(filename: str) -> pathlib.Path:
+    """Get a pathlib.Path object for a resource."""
+    assert not posixpath.isabs(filename), filename
+    assert os.path.pardir not in filename.split(posixpath.sep), filename
+
+    if hasattr(sys, 'frozen'):
+        # For PyInstaller, where we can't store resource files in a qutebrowser/ folder
+        # because the executable is already named "qutebrowser" (at least on macOS).
+        return pathlib.Path(sys.executable).parent / filename
+
+    return importlib_resources.files(qutebrowser) / filename
+
+
+def _glob_resources(
+    resource_path: pathlib.Path,
+    subdir: str,
+    ext: str,
+) -> Iterable[str]:
+    """Find resources with the given extension.
+
+    Yields a resource name like "html/log.html" (as string).
+    """
+    assert '*' not in ext, ext
+    assert ext.startswith('.'), ext
+    path = resource_path / subdir
+
+    if isinstance(resource_path, pathlib.Path):
+        for full_path in path.glob(f'*{ext}'):  # . is contained in ext
+            yield full_path.relative_to(resource_path).as_posix()
+    else:  # zipfile.Path or importlib_resources compat object
+        # Unfortunately, we can't tell mypy about resource_path being of type
+        # Union[pathlib.Path, zipfile.Path] because we set "python_version = 3.6" in
+        # .mypy.ini, but the zipfiel stubs (correctly) only declare zipfile.Path with
+        # Python 3.8...
+        assert path.is_dir(), path  # type: ignore[unreachable]
+        for subpath in path.iterdir():
+            if subpath.name.endswith(ext):
+                yield posixpath.join(subdir, subpath.name)
+
+
 def preload_resources() -> None:
     """Load resource files into the cache."""
-    for subdir, pattern in [('html', '*.html'), ('javascript', '*.js')]:
-        path = resource_filename(subdir)
-        for full_path in glob.glob(os.path.join(path, pattern)):
-            sub_path = '/'.join([subdir, os.path.basename(full_path)])
-            _resource_cache[sub_path] = read_file(sub_path)
+    resource_path = _resource_path('')
+    for subdir, ext in [('html', '.html'), ('javascript', '.js')]:
+        for name in _glob_resources(resource_path, subdir, ext):
+            _resource_cache[name] = read_file(name)
 
 
-# FIXME:typing Return value should be bytes/str
-def read_file(filename: str, binary: bool = False) -> typing.Any:
+def read_file(filename: str) -> str:
     """Get the contents of a file contained with qutebrowser.
 
     Args:
         filename: The filename to open as string.
-        binary: Whether to return a binary string.
-                If False, the data is UTF-8-decoded.
 
     Return:
         The file contents as string.
     """
-    assert not posixpath.isabs(filename), filename
-    assert os.path.pardir not in filename.split(posixpath.sep), filename
-
-    if not binary and filename in _resource_cache:
+    if filename in _resource_cache:
         return _resource_cache[filename]
 
-    if hasattr(sys, 'frozen'):
-        # PyInstaller doesn't support pkg_resources :(
-        # https://github.com/pyinstaller/pyinstaller/wiki/FAQ#misc
-        fn = os.path.join(os.path.dirname(sys.executable), filename)
-        if binary:
-            with open(fn, 'rb') as f:  # type: typing.IO
-                return f.read()
-        else:
-            with open(fn, 'r', encoding='utf-8') as f:
-                return f.read()
-    else:
-        data = pkg_resources.resource_string(
-            qutebrowser.__name__, filename)
-
-        if binary:
-            return data
-
-        return data.decode('UTF-8')
+    path = _resource_path(filename)
+    return path.read_text(encoding='utf-8')
 
 
-def resource_filename(filename: str) -> str:
-    """Get the absolute filename of a file contained with qutebrowser.
+def read_file_binary(filename: str) -> bytes:
+    """Get the contents of a binary file contained with qutebrowser.
 
     Args:
-        filename: The filename.
+        filename: The filename to open as string.
 
     Return:
-        The absolute filename.
+        The file contents as a bytes object.
     """
-    if hasattr(sys, 'frozen'):
-        return os.path.join(os.path.dirname(sys.executable), filename)
-    return pkg_resources.resource_filename(qutebrowser.__name__, filename)
+    path = _resource_path(filename)
+    return path.read_bytes()
 
 
-def _get_color_percentage(a_c1: int, a_c2: int, a_c3:
-                          int, b_c1: int, b_c2: int, b_c3: int,
-                          percent: int) -> typing.Tuple[int, int, int]:
-    """Get a color which is percent% interpolated between start and end.
-
-    Args:
-        a_c1, a_c2, a_c3: Start color components (R, G, B / H, S, V / H, S, L)
-        b_c1, b_c2, b_c3: End color components (R, G, B / H, S, V / H, S, L)
-        percent: Percentage to interpolate, 0-100.
-                 0: Start color will be returned.
-                 100: End color will be returned.
-
-    Return:
-        A (c1, c2, c3) tuple with the interpolated color components.
-    """
-    if not 0 <= percent <= 100:
-        raise ValueError("percent needs to be between 0 and 100!")
-    out_c1 = round(a_c1 + (b_c1 - a_c1) * percent / 100)
-    out_c2 = round(a_c2 + (b_c2 - a_c2) * percent / 100)
-    out_c3 = round(a_c3 + (b_c3 - a_c3) * percent / 100)
-    return (out_c1, out_c2, out_c3)
-
-
-def interpolate_color(
-        start: QColor,
-        end: QColor,
-        percent: int,
-        colorspace: typing.Optional[QColor.Spec] = QColor.Rgb
-) -> QColor:
-    """Get an interpolated color value.
-
-    Args:
-        start: The start color.
-        end: The end color.
-        percent: Which value to get (0 - 100)
-        colorspace: The desired interpolation color system,
-                    QColor::{Rgb,Hsv,Hsl} (from QColor::Spec enum)
-                    If None, start is used except when percent is 100.
-
-    Return:
-        The interpolated QColor, with the same spec as the given start color.
-    """
-    qtutils.ensure_valid(start)
-    qtutils.ensure_valid(end)
-
-    if colorspace is None:
-        if percent == 100:
-            return QColor(*end.getRgb())
-        else:
-            return QColor(*start.getRgb())
-
-    out = QColor()
-    if colorspace == QColor.Rgb:
-        a_c1, a_c2, a_c3, _alpha = start.getRgb()
-        b_c1, b_c2, b_c3, _alpha = end.getRgb()
-        components = _get_color_percentage(a_c1, a_c2, a_c3, b_c1, b_c2, b_c3,
-                                           percent)
-        out.setRgb(*components)
-    elif colorspace == QColor.Hsv:
-        a_c1, a_c2, a_c3, _alpha = start.getHsv()
-        b_c1, b_c2, b_c3, _alpha = end.getHsv()
-        components = _get_color_percentage(a_c1, a_c2, a_c3, b_c1, b_c2, b_c3,
-                                           percent)
-        out.setHsv(*components)
-    elif colorspace == QColor.Hsl:
-        a_c1, a_c2, a_c3, _alpha = start.getHsl()
-        b_c1, b_c2, b_c3, _alpha = end.getHsl()
-        components = _get_color_percentage(a_c1, a_c2, a_c3, b_c1, b_c2, b_c3,
-                                           percent)
-        out.setHsl(*components)
-    else:
-        raise ValueError("Invalid colorspace!")
-    out = out.convertTo(start.spec())
-    qtutils.ensure_valid(out)
-    return out
+def parse_version(version: str) -> VersionNumber:
+    """Parse a version string."""
+    v_q, _suffix = QVersionNumber.fromString(version)
+    return cast(VersionNumber, v_q.normalized())
 
 
 def format_seconds(total_seconds: int) -> str:
@@ -303,12 +278,10 @@ def format_seconds(total_seconds: int) -> str:
     return prefix + ':'.join(chunks)
 
 
-def format_size(size: typing.Optional[float],
-                base: int = 1024,
-                suffix: str = '') -> str:
+def format_size(size: Optional[float], base: int = 1024, suffix: str = '') -> str:
     """Format a byte size so it's human readable.
 
-    Inspired by http://stackoverflow.com/q/1094841
+    Inspired by https://stackoverflow.com/q/1094841
     """
     prefixes = ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
     if size is None:
@@ -324,13 +297,13 @@ class FakeIOStream(io.TextIOBase):
 
     """A fake file-like stream which calls a function for write-calls."""
 
-    def __init__(self, write_func: typing.Callable[[str], int]) -> None:
+    def __init__(self, write_func: Callable[[str], int]) -> None:
         super().__init__()
         self.write = write_func  # type: ignore[assignment]
 
 
 @contextlib.contextmanager
-def fake_io(write_func: typing.Callable[[str], int]) -> typing.Iterator[None]:
+def fake_io(write_func: Callable[[str], int]) -> Iterator[None]:
     """Run code with stdout and stderr replaced by FakeIOStreams.
 
     Args:
@@ -354,7 +327,7 @@ def fake_io(write_func: typing.Callable[[str], int]) -> typing.Iterator[None]:
 
 
 @contextlib.contextmanager
-def disabled_excepthook() -> typing.Iterator[None]:
+def disabled_excepthook() -> Iterator[None]:
     """Run code with the exception hook temporarily disabled."""
     old_excepthook = sys.excepthook
     sys.excepthook = sys.__excepthook__
@@ -387,7 +360,7 @@ class prevent_exceptions:  # noqa: N801,N806 pylint: disable=invalid-name
         _predicate: The condition which needs to be True to prevent exceptions
     """
 
-    def __init__(self, retval: typing.Any, predicate: bool = True) -> None:
+    def __init__(self, retval: Any, predicate: bool = True) -> None:
         """Save decorator arguments.
 
         Gets called on parse-time with the decorator arguments.
@@ -398,7 +371,7 @@ class prevent_exceptions:  # noqa: N801,N806 pylint: disable=invalid-name
         self._retval = retval
         self._predicate = predicate
 
-    def __call__(self, func: typing.Callable) -> typing.Callable:
+    def __call__(self, func: Callable) -> Callable:
         """Called when a function should be decorated.
 
         Args:
@@ -413,7 +386,7 @@ class prevent_exceptions:  # noqa: N801,N806 pylint: disable=invalid-name
         retval = self._retval
 
         @functools.wraps(func)
-        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
+        def wrapper(*args: Any, **kwargs: Any) -> Any:
             """Call the original function."""
             try:
                 return func(*args, **kwargs)
@@ -424,7 +397,7 @@ class prevent_exceptions:  # noqa: N801,N806 pylint: disable=invalid-name
         return wrapper
 
 
-def is_enum(obj: typing.Any) -> bool:
+def is_enum(obj: Any) -> bool:
     """Check if a given object is an enum."""
     try:
         return issubclass(obj, enum.Enum)
@@ -432,9 +405,7 @@ def is_enum(obj: typing.Any) -> bool:
         return False
 
 
-def get_repr(obj: typing.Any,
-             constructor: bool = False,
-             **attrs: typing.Any) -> str:
+def get_repr(obj: Any, constructor: bool = False, **attrs: Any) -> str:
     """Get a suitable __repr__ string for an object.
 
     Args:
@@ -457,7 +428,7 @@ def get_repr(obj: typing.Any,
             return '<{}>'.format(cls)
 
 
-def qualname(obj: typing.Any) -> str:
+def qualname(obj: Any) -> str:
     """Get the fully qualified name of an object.
 
     Based on twisted.python.reflect.fullyQualifiedName.
@@ -485,14 +456,10 @@ def qualname(obj: typing.Any) -> str:
         return repr(obj)
 
 
-# The string annotation is a WORKAROUND for a Python 3.5.2 bug:
-# https://github.com/python/typing/issues/266
+_ExceptionType = Union[Type[BaseException], Tuple[Type[BaseException]]]
 
-def raises(exc: ('typing.Union['  # pylint: disable=bad-docstring-quotes
-                 '    typing.Type[BaseException], '
-                 '    typing.Tuple[typing.Type[BaseException]]]'),
-           func: typing.Callable,
-           *args: typing.Any) -> bool:
+
+def raises(exc: _ExceptionType, func: Callable, *args: Any) -> bool:
     """Check if a function raises a given exception.
 
     Args:
@@ -520,7 +487,7 @@ def force_encoding(text: str, encoding: str) -> str:
 
 
 def sanitize_filename(name: str,
-                      replacement: typing.Optional[str] = '_',
+                      replacement: Optional[str] = '_',
                       shorten: bool = False) -> str:
     """Replace invalid filename characters.
 
@@ -641,15 +608,6 @@ def supports_selection() -> bool:
     return QApplication.clipboard().supportsSelection()
 
 
-def random_port() -> int:
-    """Get a random free port."""
-    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-    sock.bind(('localhost', 0))
-    port = sock.getsockname()[1]
-    sock.close()
-    return port
-
-
 def open_file(filename: str, cmdline: str = None) -> None:
     """Open the given file.
 
@@ -708,7 +666,7 @@ def open_file(filename: str, cmdline: str = None) -> None:
     proc.start_detached(cmd, args)
 
 
-def unused(_arg: typing.Any) -> None:
+def unused(_arg: Any) -> None:
     """Function which does nothing to avoid pylint complaining."""
 
 
@@ -730,12 +688,12 @@ def expand_windows_drive(path: str) -> str:
         return path
 
 
-def yaml_load(f: typing.Union[str, typing.IO[str]]) -> typing.Any:
+def yaml_load(f: Union[str, IO[str]]) -> Any:
     """Wrapper over yaml.load using the C loader if possible."""
     start = datetime.datetime.now()
 
     # WORKAROUND for https://github.com/yaml/pyyaml/pull/181
-    with log.ignore_py_warnings(
+    with log.py_warning_filter(
             category=DeprecationWarning,
             message=r"Using or importing the ABCs from 'collections' instead "
             r"of from 'collections\.abc' is deprecated.*"):
@@ -766,8 +724,7 @@ def yaml_load(f: typing.Union[str, typing.IO[str]]) -> typing.Any:
     return data
 
 
-def yaml_dump(data: typing.Any,
-              f: typing.IO[str] = None) -> typing.Optional[str]:
+def yaml_dump(data: Any, f: IO[str] = None) -> Optional[str]:
     """Wrapper over yaml.dump using the C dumper if possible.
 
     Also returns a str instead of bytes.
@@ -780,7 +737,7 @@ def yaml_dump(data: typing.Any,
         return yaml_data.decode('utf-8')
 
 
-def chunk(elems: typing.Sequence, n: int) -> typing.Iterator[typing.Sequence]:
+def chunk(elems: Sequence, n: int) -> Iterator[Sequence]:
     """Yield successive n-sized chunks from elems.
 
     If elems % n != 0, the last chunk will be smaller.
@@ -833,3 +790,81 @@ def libgl_workaround() -> None:
     libgl = ctypes.util.find_library("GL")
     if libgl is not None:  # pragma: no branch
         ctypes.CDLL(libgl, mode=ctypes.RTLD_GLOBAL)
+
+
+def parse_duration(duration: str) -> int:
+    """Parse duration in format XhYmZs into milliseconds duration."""
+    if duration.isdigit():
+        # For backward compatibility return milliseconds
+        return int(duration)
+
+    match = re.fullmatch(
+        r'(?P<hours>[0-9]+(\.[0-9])?h)?\s*'
+        r'(?P<minutes>[0-9]+(\.[0-9])?m)?\s*'
+        r'(?P<seconds>[0-9]+(\.[0-9])?s)?',
+        duration
+    )
+    if not match or not match.group(0):
+        raise ValueError(
+            f"Invalid duration: {duration} - "
+            "expected XhYmZs or a number of milliseconds"
+        )
+    seconds_string = match.group('seconds') if match.group('seconds') else '0'
+    seconds = float(seconds_string.rstrip('s'))
+    minutes_string = match.group('minutes') if match.group('minutes') else '0'
+    minutes = float(minutes_string.rstrip('m'))
+    hours_string = match.group('hours') if match.group('hours') else '0'
+    hours = float(hours_string.rstrip('h'))
+    milliseconds = int((seconds + minutes * 60 + hours * 3600) * 1000)
+    return milliseconds
+
+
+def mimetype_extension(mimetype: str) -> Optional[str]:
+    """Get a suitable extension for a given mimetype.
+
+    This mostly delegates to Python's mimetypes.guess_extension(), but backports some
+    changes (via a simple override dict) which are missing from earlier Python versions.
+    Most likely, this can be dropped once the minimum Python version is raised to 3.7.
+    """
+    overrides = {
+        # Added around 3.8
+        "application/manifest+json": ".webmanifest",
+        "application/x-hdf5": ".h5",
+
+        # Added in Python 3.7
+        "application/wasm": ".wasm",
+
+        # Wrong values for Python 3.6
+        # https://bugs.python.org/issue1043134
+        # https://github.com/python/cpython/pull/14375
+        "application/octet-stream": ".bin",  # not .a
+        "application/postscript": ".ps",  # not .ai
+        "application/vnd.ms-excel": ".xls",  # not .xlb
+        "application/vnd.ms-powerpoint": ".ppt",  # not .pot
+        "application/xml": ".xsl",  # not .rdf
+        "audio/mpeg": ".mp3",  # not .mp2
+        "image/jpeg": ".jpg",  # not .jpe
+        "image/tiff": ".tiff",  # not .tif
+        "text/html": ".html",  # not .htm
+        "text/plain": ".txt",  # not .bat
+        "video/mpeg": ".mpeg",  # not .m1v
+    }
+    if mimetype in overrides:
+        return overrides[mimetype]
+    return mimetypes.guess_extension(mimetype, strict=False)
+
+
+@contextlib.contextmanager
+def cleanup_file(filepath: str) -> Iterator[None]:
+    """Context that deletes a file upon exit or error.
+
+    Args:
+        filepath: The file path
+    """
+    try:
+        yield
+    finally:
+        try:
+            os.remove(filepath)
+        except OSError as e:
+            log.misc.error(f"Failed to delete tempfile {filepath} ({e})!")
